#!/usr/bin/python
# -*- coding: UTF-8 -*-

# Author Cleoner S. Pietralonga
# e-mail: cleonerp@gmail.com
# Apache License

import numpy as np
from sympy.physics.quantum import TensorProduct

cupy_is_available = True
try:
    import cupy as cp
except Exception as ex:
    cupy_is_available = False
    print("Cuda is not available!")

from logicqubit.utils import *

"""
Hilbert space
"""
class Hilbert():
    __number_of_qubits = 1
    __cuda = True
    __first_left = True

    @staticmethod
    def ket(value):  # get ket state
        result = Matrix([[Utils.onehot(i, value)] for i in range(2)], Hilbert.__cuda)
        return result

    @staticmethod
    def bra(value):  # get bra state
        result = Matrix([Utils.onehot(i, value) for i in range(2)], Hilbert.__cuda)
        return result

    @staticmethod
    def getState():  # get state of all qubits
        if Hilbert.getCuda():
            state = Hilbert.kronProduct([Hilbert.ket(0) for i in range(Hilbert.getNumberOfQubits())])
        else:
            if Hilbert.isFirstLeft():
                a = sp.symbols([str(i) + "a" + str(i) + "_0" for i in range(1, Hilbert.getNumberOfQubits() + 1)])
                b = sp.symbols([str(i) + "b" + str(i) + "_1" for i in range(1, Hilbert.getNumberOfQubits() + 1)])
            else:
                a = sp.symbols([str(Hilbert.getNumberOfQubits() + 1 - i) + "a" + str(i) + "_0" for i in
                                reversed(range(1, Hilbert.getNumberOfQubits() + 1))])
                b = sp.symbols([str(Hilbert.getNumberOfQubits() + 1 - i) + "b" + str(i) + "_1" for i in
                                reversed(range(1, Hilbert.getNumberOfQubits() + 1))])
            state = Hilbert.kronProduct([Hilbert.ket(0) * a[i] + Hilbert.ket(1) * b[i] for i in range(Hilbert.getNumberOfQubits())])
        return state

    @staticmethod
    def getAdjoint(psi):  # get adjoint matrix
        result = psi.adjoint()
        return result

    @staticmethod
    def product(Operator, psi):  # performs an operation between the operator and the psi state
        result = Operator * psi
        return result

    @staticmethod
    def kronProduct(list):  # Kronecker product
        A = list[0]  # acts in qubit 1 which is the left most
        for M in list[1:]:
            A = A.kron(M)
        return A

    @staticmethod
    def setNumberOfQubits(number):
        Hilbert.__number_of_qubits = number

    @staticmethod
    def getNumberOfQubits():
        return Hilbert.__number_of_qubits

    @staticmethod
    def setCuda(cuda):
        Hilbert.__cuda = cuda

    @staticmethod
    def getCuda():
        return Hilbert.__cuda

    @staticmethod
    def setFirstLeft(value):
        Hilbert.__first_left = value

    @staticmethod
    def isFirstLeft():
        return Hilbert.__first_left


"""
Wrap methods from the numpy, cupy and sympy libraries.
"""
class Matrix:

    def __init__(self, matrix, cuda=True):
        self.__matrix = matrix
        self.__cuda = cuda
        if isinstance(matrix, list):  # if it's a list
            if self.__cuda:
                if cupy_is_available:
                    self.__matrix = cp.array(matrix)  # create matrix with cupy
                else:
                    self.__matrix = np.array(matrix)  # create matrix with numpy
            else:
                self.__matrix = sp.Matrix(matrix)  # create matrix with sympy
        else:
            if isinstance(matrix, Matrix):  # if it's a Matrix class
                self.__matrix = matrix.get()
            else:
                self.__matrix = matrix

    def __add__(self, other):  # sum of the matrices
        result = self.__matrix + other.get()
        return Matrix(result, self.__cuda)

    def __sub__(self, other):  # subtraction of the matrices
        result = self.__matrix - other.get()
        return Matrix(result, self.__cuda)

    def __mul__(self, other):  # product of the matrices
        if isinstance(other, Matrix):
            other = other.get()
            if self.__cuda:
                if cupy_is_available:
                    result = cp.dot(self.__matrix, other)  # for cupy matrix
                else:
                    result = np.dot(self.__matrix, other)  # for numpy matrix
            else:
                result = self.__matrix * other
        else:
            result = self.__matrix * other
        return Matrix(result, self.__cuda)

    def __truediv__(self, other):
        result = self.__matrix * (1./other)
        return Matrix(result, self.__cuda)

    def __eq__(self, other):
        return self.__matrix == other.get()

    def __str__(self):
        return str(self.__matrix)

    def kron(self, other):  # Kronecker product
        if self.__cuda:
            if cupy_is_available:
                result = cp.kron(self.__matrix, other.get())
            else:
                result = np.kron(self.__matrix, other.get())
        else:
            result = TensorProduct(self.__matrix, other.get())
        return Matrix(result, self.__cuda)

    def get(self):
        return self.__matrix

    def getAngles(self):  # converts state coefficients into angles
        angles = []
        if self.__cuda:
            if cupy_is_available:
                angles = cp.angle(self.__matrix)
            else:
                angles = np.angle(self.__matrix)
        else:
            print("This session is symbolic!")
        return angles

    def trace(self):  # get matrix trace
        result = self.__matrix.trace()
        return Matrix(result, self.__cuda)

    def adjoint(self):  # get matrix adjoint
        if self.__cuda:
            result = self.__matrix.transpose().conj()
        else:
            result = self.__matrix.transpose().conjugate()
        return Matrix(result, self.__cuda)



#!/usr/bin/python
# -*- coding: UTF-8 -*-

# Author Cleoner S. Pietralonga
# e-mail: cleonerp@gmail.com
# Apache License

import random
import matplotlib.pyplot as plt
from matplotlib import cm

from logicqubit.qubits import *
from logicqubit.oracle import *
from logicqubit.utils import *

"""
Main class that allows you to create the circuit.
In this class the quantum gates methods perform and record the type of operation.
It is necessary to enter the qubit id as an input parameter.
"""
class LogicQuBit(Qubits, Gates, Circuit):

    def __init__(self, number_of_qubits = 3, **kwargs):
        symbolic = kwargs.get('symbolic', False)
        first_left = kwargs.get('first_left', True)  # qubit 1 is the left most
        super().setCuda(not symbolic)
        super().setFirstLeft(first_left)
        super().setNumberOfQubits(number_of_qubits)
        Qubits.__init__(self)
        Gates.__init__(self, number_of_qubits)
        Circuit.__init__(self)

    # One qubit gates
    # input parameters: target
    # .......................................
    def X(self, target):
        self.addOp("X", self.qubitsToList([target]))
        operator = super().X(target)
        self.setOperation(operator)

    def Y(self, target):
        self.addOp("Y", self.qubitsToList([target]))
        operator = super().Y(target)
        self.setOperation(operator)

    def Z(self, target):
        self.addOp("Z", self.qubitsToList([target]))
        operator = super().Z(target)
        self.setOperation(operator)

    def V(self, target):
        self.addOp("V", self.qubitsToList([target]))
        operator = super().V(target)
        self.setOperation(operator)

    def S(self, target):
        self.addOp("S", self.qubitsToList([target]))
        operator = super().S(target)
        self.setOperation(operator)

    def T(self, target):
        self.addOp("T", self.qubitsToList([target]))
        operator = super().T(target)
        self.setOperation(operator)

    def H(self, target):
        self.addOp("H", self.qubitsToList([target]))
        operator = super().H(target)
        self.setOperation(operator)

    def U(self, target, *argv):
        self.addOp("U", self.qubitsToList([target, argv]))
        operator = super().U(target, argv)
        self.setOperation(operator)

    def U3(self, target, theta, phi, _lambda):
        self.addOp("U3", self.qubitsToList([target, theta, phi, _lambda]))
        operator = super().U3(target, theta, phi, _lambda)
        self.setOperation(operator)

    def U2(self, target, phi, _lambda):
        self.addOp("U2", self.qubitsToList([target, phi, _lambda]))
        operator = super().U2(target, phi, _lambda)
        self.setOperation(operator)

    def U1(self, target, _lambda):
        self.addOp("U1", self.qubitsToList([target, _lambda]))
        operator = super().U1(target, _lambda)
        self.setOperation(operator)

    def RX(self, target, theta):
        self.addOp("RX", self.qubitsToList([target, theta]))
        operator = super().RX(target, theta)
        self.setOperation(operator)

    def RY(self, target, theta):
        self.addOp("RY", self.qubitsToList([target, theta]))
        operator = super().RY(target, theta)
        self.setOperation(operator)

    def RZ(self, target, phi):
        self.addOp("RZ", self.qubitsToList([target, phi]))
        operator = super().RZ(target, phi)
        self.setOperation(operator)

    # Two qubit gates
    # input parameters: control and target
    # .......................................
    def CH(self, control, target):
        self.addOp("CH", self.qubitsToList([control, target]))
        operator = super().CH(control, target)
        self.setOperation(operator)

    def CX(self, control, target):
        self.addOp("CX", self.qubitsToList([control, target]))
        operator = super().CX(control, target)
        self.setOperation(operator)

    def CNOT(self, control, target):
        self.CX(control, target)

    def CY(self, control, target):
        self.addOp("CY", self.qubitsToList([control, target]))
        operator = super().CY(control, target)
        self.setOperation(operator)

    def CZ(self, control, target):
        self.addOp("CZ", self.qubitsToList([control, target]))
        operator = super().CZ(control, target)
        self.setOperation(operator)

    def CRX(self, control, target, theta):
        self.addOp("CRX", self.qubitsToList([control, target, theta]))
        operator = super().CRX(control, target, theta)
        self.setOperation(operator)

    def CRY(self, control, target, theta):
        self.addOp("CRY", self.qubitsToList([control, target, theta]))
        operator = super().CRY(control, target, theta)
        self.setOperation(operator)

    def CRZ(self, control, target, phi):
        self.addOp("CRZ", self.qubitsToList([control, target, phi]))
        operator = super().CRZ(control, target, phi)
        self.setOperation(operator)

    def CU(self, control, target, *argv):
        self.addOp("CU", self.qubitsToList([control, target]))
        operator = super().CU(control, target, argv)
        self.setOperation(operator)

    def CU3(self, control, target, theta, phi, _lambda):
        self.addOp("CU3", self.qubitsToList([control, target, theta, phi, _lambda]))
        operator = super().CU3(control, target, theta, phi, _lambda)
        self.setOperation(operator)

    def CU2(self, control, target, phi, _lambda):
        self.addOp("CU2", self.qubitsToList([control, target, phi, _lambda]))
        operator = super().CU2(control, target, phi, _lambda)
        self.setOperation(operator)

    def CU1(self, control, target, _lambda):
        self.addOp("CU1", self.qubitsToList([control, target, _lambda]))
        operator = super().CU1(control, target, _lambda)
        self.setOperation(operator)

    # Three qubit gates
    # input parameters: control1, control2, and target
    # .......................................
    def CCX(self, control1, control2, target):
        self.addOp("CCX", self.qubitsToList([control1, control2, target]))
        operator = super().CCX(control1, control2, target)
        self.setOperation(operator)

    def Toffoli(self, control1, control2, target):
        self.CCX(control1, control2, target)

    def SWAP(self, target1, target2):
        self.addOp("SWAP", self.qubitsToList([target1, target2]))
        operator = super().SWAP(target1, target2)
        self.setOperation(operator)

    def Fredkin(self, control, target1, target2):
        self.addOp("Fredkin", self.qubitsToList([control, target1, target2]))
        operator = super().Fredkin(control, target1, target2)
        self.setOperation(operator)

    # adds oracle through Zhegalkin polynomials
    def addOracle(self, oracle):
        targets, input_qubits, truth_table = oracle.get()
        poly = Zhegalkin_Poly()
        for bits in truth_table:
            poly.addTable(bits)
        result = poly.Compute()
        for target, p in enumerate(result):
            for i, value in reversed(list(enumerate(p))):
                if(value==1):
                    blist = [int(i, base=2) for i in bin(i)[2:].zfill(len(input_qubits))]
                    if(bin(i).count("1") == 2):  # use Toffoli gate
                        try:
                            q1 = blist.index(1)
                            q2 = blist[q1+1:].index(1)+q1+1
                            self.CCX(input_qubits[q1], input_qubits[q2], targets[target])
                        except Exception as e:
                            print(e)
                    elif(bin(i).count("1")==1):  # use controlled not gate
                        try:
                            q = blist.index(1)
                            self.CX(input_qubits[q], targets[target])
                        except Exception as e:
                            print(e)
                    elif(i==0):  # use not gate
                        self.X(targets[target])

    # calculate the density matrix
    def DensityMatrix(self):
        density_m = self.getPsi() * self.getPsiAdjoint()
        return density_m

    # purity of state
    def Pure(self):
        density_m = self.DensityMatrix()
        pure = (density_m*density_m).trace()
        return pure

    # choose a state among many others
    def get_shot(self, measured, shots):
        max_set = shots*100
        if not self.getCuda():
            list_all = [int(value * max_set) * [i] for i, value in enumerate(measured)]
        else:
            list_all = [int(value.real * max_set) * [i] for i, value in enumerate(measured)]
        list_all_values = []
        for list_values in list_all:
            list_all_values += list_values
        values = [random.choice(list_all_values) for i in range(shots)]  # choose a state
        return values

    # measure only one qubit
    def Measure_One(self, target, shots=1):
        self.addOp("Measure", self.qubitsToList([target]))
        density_m = self.DensityMatrix()
        list = self.getOrdListSimpleGate(target, super().P0())
        P0 = self.kronProduct(list)
        list = self.getOrdListSimpleGate(target, super().P1())
        P1 = self.kronProduct(list)
        measure_0 = (density_m*P0).trace().get()  # expected value for 0
        measure_1 = (density_m*P1).trace().get()  # expected value for 1
        value = self.get_shot([measure_0, measure_1], shots)  # measure values without averaging
        if(value[0] == 0):
            new_state = self.product(P0, self.getPsi())/sqrt(measure_0)  # collapse of the state after the measure
        else:
            new_state = self.product(P1, self.getPsi())/sqrt(measure_1)  # collapse of the state after the measure
        self.setPsi(new_state)
        return value

    # measure a list of qubits
    def Measure(self, target, fisrt_msb = False):  # ex: measure 3 qubits of 5: 2,1,4 of state "010" -> M010 = |1><1| x |0><0| x 1 x |0><0| x 1
        target = self.qubitsToList(target)
        if(fisrt_msb):  # if fisrt_msb=True -> the first on the list will be the most significant
            target.reverse()
        self.setMeasuredQubits(target)
        self.addOp("Measure", target)
        density_m = self.DensityMatrix()
        size_p = len(target)  # number of qubits to be measured
        size = 2 ** size_p  # number of possible states
        result = []
        for i in range(size):
            tlist = [self.ID() for tl in range(self.getNumberOfQubits())]
            blist = [i >> bl & 0x1 for bl in range(size_p)]  # bit list, bits de cada i
            for j, value in enumerate(target):
                if blist[j] == 0:  # most significant first
                    tlist[value-1] = super().P0()
                else:
                    tlist[value-1] = super().P1()
            if not self.isFirstLeft():
                tlist.reverse()
            M = self.kronProduct(tlist)
            measure = (density_m * M).get().trace()  # expected value
            if self.getCuda() and cupy_is_available:
                measure = measure.get().item().real
            result.append(measure)
        self.setMeasuredValues(result)
        return result

    # plot state probability distribution
    def Plot(self, big_endian=False):
        size_p = len(self.getMeasuredQubits())  # number of bits measured
        if(size_p > 0):
            size = 2 ** size_p
            if(big_endian):
                names = ["|" + ''.join(list(reversed("{0:b}".format(i).zfill(size_p)))) + ">" for i in range(size)]
            else:
                names = ["|" + "{0:b}".format(i).zfill(size_p) + ">" for i in range(size)]
            values = self.getMeasuredValues()
            #plt.figure(figsize = (5, 3))
            plt.bar(names, values)
            plt.xticks(rotation=50)
            plt.suptitle('')
            plt.show()
        else:
            print("No qubit measured!")

    # plot the density matrix
    def PlotDensityMatrix(self, imaginary=False, decimal=False):
        size_p = self.getNumberOfQubits()  # number of qubits
        mRho = [[0]*2**size_p for i in range(2**size_p)]
        rho = self.DensityMatrix().get()
        for id1 in range(2**size_p):
            for id2 in range(2**size_p):
                if self.getCuda():
                    value = rho[id1][id2]
                    if(not imaginary):
                        value = value.item().real
                    else:
                        value = value.item().imag
                else:
                    value = rho[id1,id2]
                    if(not imaginary):
                        value = sp.re(value)
                    else:
                        value = sp.im(value)
                mRho[id1][id2] = value

        result = np.array(mRho, dtype=float)
        fig = plt.figure(figsize=(5, 5), dpi=150)
        ax1 = fig.add_subplot(111, projection='3d')

        size = 2 ** size_p
        if(not decimal):
            labels = ["|" + "{0:b}".format(i).zfill(size_p) + ">" for i in range(size)]
        else:
            labels = [i for i in range(size)]
        xlabels = np.array(labels)
        xpos = np.arange(xlabels.shape[0])
        ylabels = np.array(labels)
        ypos = np.arange(ylabels.shape[0])

        xpos_mesh, ypos_mesh = np.meshgrid(xpos, ypos, copy=False)

        zpos = result
        zpos = zpos.ravel()

        dx = 0.5
        dy = 0.5
        dz = zpos

        ax1.set_xticks(xpos + dx / 2.)
        ax1.set_xticklabels(xlabels)
        ax1.set_yticks(ypos + dy / 2.)
        ax1.set_yticklabels(ylabels)

        values = np.linspace(0.2, 1., xpos_mesh.ravel().shape[0])
        colors = cm.rainbow(values)
        ax1.bar3d(xpos_mesh.ravel(), ypos_mesh.ravel(), dz * 0, dx, dy, dz, color=colors)
        plt.show()



#!/usr/bin/python
# -*- coding: UTF-8 -*-

# Author Cleoner S. Pietralonga
# e-mail: cleonerp@gmail.com
# Apache License

from IPython.display import display, Math, Latex

from logicqubit.hilbert import *
from logicqubit.gates import *
from logicqubit.circuit import *
from logicqubit.utils import *

class Qubits(Hilbert):

    def __init__(self):
        Qubits.__number = 0
        Qubits.__used_qubits = []
        Qubits.__measured_qubits = []
        Qubits.__measured_values = []
        if self.getCuda():
            Qubits.__psi = self.kronProduct([self.ket(0) for i in range(self.getNumberOfQubits())])
        else:
            if self.isFirstLeft():  # qubit 1 is the first on the left
                a = sp.symbols([str(i) + "a" + str(i) + "_0" for i in range(1, self.getNumberOfQubits() + 1)])
                b = sp.symbols([str(i) + "b" + str(i) + "_1" for i in range(1, self.getNumberOfQubits() + 1)])
            else:
                a = sp.symbols([str(self.getNumberOfQubits()+1-i) + "a" + str(i) + "_0" for i in reversed(range(1, self.getNumberOfQubits() + 1))])
                b = sp.symbols([str(self.getNumberOfQubits()+1-i) + "b" + str(i) + "_1" for i in reversed(range(1, self.getNumberOfQubits() + 1))])
            Qubits.__psi = self.kronProduct([self.ket(0)*a[i]+self.ket(1)*b[i] for i in range(self.getNumberOfQubits())])

    def addQubit(self, id=None):
        if(len(Qubits.__used_qubits) < self.getNumberOfQubits()):
            if(id != None):
                if(not id in Qubits.__used_qubits):
                    Qubits.__used_qubits.append(id)
                else:
                    print("Qubit already used!")
            else:
                id = self.getLowestIdAvailable()
                Qubits.__used_qubits.append(id)
            return id
        else:
            print("All qubits have already been used!")
            return None

    def getLowestIdAvailable(self):
        all = list(range(1, self.getNumberOfQubits() + 1))
        for i in Qubits.__used_qubits:
            all.remove(i)
        return min(all)

    def getBiggestIdAvailable(self):
        all = list(range(1, self.getNumberOfQubits() + 1))
        for i in Qubits.__used_qubits:
            all.remove(i)
        return max(all)

    def getQubitsNumber(self):
        return self.getNumberOfQubits()

    def getUsedQubits(self):
        return len(Qubits.__used_qubits)

    def setMeasuredQubits(self, target):
        if(isinstance(target, list)):
            for id in target:
                Qubits.__measured_qubits.append(id)
        else:
            Qubits.__measured_qubits.append(target)

    def getMeasuredQubits(self):
        return Qubits.__measured_qubits

    def isMeasured(self, target):
        if (isinstance(target, list)):
            for id in target:
                if(id in Qubits.__measured_qubits):
                    print("Qubit "+str(id)+" alredy measured!")
                    return True
        else:
            if(target in Qubits.__measured_qubits):
                return True
        return False

    def setMeasuredValues(self, value):
        Qubits.__measured_values = value

    def getMeasuredValues(self):
        return Qubits.__measured_values

    def setPsi(self, psi):
        Qubits.__psi = psi

    def getPsi(self):
        return Qubits.__psi

    def getDictPsi(self):
        size_p = self.getQubitsNumber()
        size = 2 ** size_p
        labels = ["{0:b}".format(i).zfill(size_p) for i in range(size)]
        if not self.getCuda():
            value_l = [Utils.textSymbolfix(str(value), self.getQubitsNumber(), self.isFirstLeft()) for value in Qubits.__psi.get()]
            dictPsi = {label: value_l[i] for i, label in enumerate(labels)}
        else:
            dictPsi = {label: Qubits.__psi.get()[i].item() for i, label in enumerate(labels)}
        return dictPsi

    def getPsiAtAngles(self, degree=False):
        angles = Qubits.__psi.getAngles()
        if(degree):
            angles = angles*180/pi
            angles = (angles + 360)%360
        return angles

    def getPsiAdjoint(self):
        result = Qubits.__psi.adjoint()
        return result

    def setOperation(self, operator):
        Qubits.__psi = self.product(operator, Qubits.__psi)
        Qubits.__last_operator = operator

    def qubitsToList(self, values):
        result = []
        for value in values:
            if (isinstance(value, Qubit)):
                result.append(value.getId())
            else:
                result.append(value)
        return result

    def setSymbolValuesForAll(self, a, b):  # replace all symbols with real values
        if not self.getCuda():
            for i in range(1, self.getNumberOfQubits()+1):
                if (self.isFirstLeft()):
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"a"+str(i)+"_0", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"a"+str(i)+"_1", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"b"+str(i)+"_0", b), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"b"+str(i)+"_1", b), False)
                else:
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"a"+str(i)+"_0", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"a"+str(i)+"_1", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"b"+str(i)+"_0", b), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"b"+str(i)+"_1", b), False)
        else:
            print("This session is not symbolic!")

    def setSymbolValuesForListId(self, id, a, b):  # replace symbols with real values using a list of qubit ids
        if not self.getCuda():
            list_id = self.qubitsToList(id)
            for i in list_id:
                if (self.isFirstLeft()):
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"a"+str(i)+"_0", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"a"+str(i)+"_1", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"b"+str(i)+"_0", b), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+"b"+str(i)+"_1", b), False)
                else:
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"a"+str(i)+"_0", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"a"+str(i)+"_1", a), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"b"+str(i)+"_0", b), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+"b"+str(i)+"_1", b), False)
        else:
            print("This session is not symbolic!")

    def setSymbolValue(self, id, symbol, value):  # replace symbols with real values by id of qubit
        if not self.getCuda():
            list_id = self.qubitsToList(id)
            for i in list_id:
                if self.isFirstLeft():
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+symbol+str(i)+"_0", value), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(i)+symbol+str(i)+"_1", value), False)
                else:
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+symbol+str(i)+"_0", value), False)
                    Qubits.__psi = Matrix(Qubits.__psi.get().subs(str(self.getNumberOfQubits()+1-i)+symbol+str(i)+"_1", value), False)
        else:
            print("This session is not symbolic!")

    def PrintState(self, simple = False):
        if self.getCuda():
            value = Utils.vec2tex(Qubits.__psi.get())
        else:
            value = Utils.texfix(Qubits.__psi.get(), self.getNumberOfQubits(), self.isFirstLeft())

        if(not simple):
            display(Math(value))
        else:
            print(value)

    def PrintLastOperator(self, tex = True):
        if(tex):
            value = sp.latex(Qubits.__last_operator.get())
            display(Math(value))
        else:
            print(Qubits.__last_operator.get())


"""
Allows to handle qubits without having to pass the id.
In this class the quantum gates methods perform and record the type of operation
"""
class Qubit(Qubits, Gates, Circuit):
    def __init__(self, id = None):
        self.__id = self.addQubit(id)
        self.__name = "q"+str(self.__id)
        Gates.__init__(self, self.getQubitsNumber())

    def __eq__(self, other):
        return self.__id == other

    def setName(self, name):
        self.__name = name

    def getName(self):
        return self.__name

    def getId(self):
        return self.__id

    def setSymbolValues(self, a, b):
        self.setSymbolValuesForListId([self.__id], a, b)

    # One qubit gates
    # .......................................
    def X(self):
        self.addOp("X", [self.__id])
        operator = super().X(self.__id)
        self.setOperation(operator)

    def Y(self):
        self.addOp("Y", [self.__id])
        operator = super().Y(self.__id)
        self.setOperation(operator)

    def Z(self):
        self.addOp("Z", [self.__id])
        operator = super().Z(self.__id)
        self.setOperation(operator)

    def V(self, adjoint = False):
        self.addOp("V", [self.__id])
        operator = super().V(self.__id, adjoint)
        self.setOperation(operator)

    def S(self, adjoint = False):
        self.addOp("S", [self.__id])
        operator = super().S(self.__id, adjoint)
        self.setOperation(operator)

    def T(self, adjoint = False):
        self.addOp("T", [self.__id])
        operator = super().T(self.__id, adjoint)
        self.setOperation(operator)

    def H(self):
        self.addOp("H", [self.__id])
        operator = super().H(self.__id)
        self.setOperation(operator)

    def U(self, *argv):
        self.addOp("U", self.qubitsToList([self.__id, argv]))
        operator = super().U(self.__id, argv)
        self.setOperation(operator)

    def U3(self, theta, phi, _lambda):
        self.addOp("U3", self.qubitsToList([self.__id, theta, phi, _lambda]))
        operator = super().U3(self.__id, theta, phi, _lambda)
        self.setOperation(operator)

    def U2(self, phi, _lambda):
        self.addOp("U2", self.qubitsToList([self.__id, phi, _lambda]))
        operator = super().U2(self.__id, phi, _lambda)
        self.setOperation(operator)

    def U1(self, _lambda):
        self.addOp("U1", self.qubitsToList([self.__id, _lambda]))
        operator = super().U1(self.__id, _lambda)
        self.setOperation(operator)

    def RX(self, theta):
        self.addOp("RX", self.qubitsToList([self.__id, theta]))
        operator = super().RX(self.__id, theta)
        self.setOperation(operator)

    def RY(self, theta):
        self.addOp("RY", self.qubitsToList([self.__id, theta]))
        operator = super().RY(self.__id, theta)
        self.setOperation(operator)

    def RZ(self, phi):
        self.addOp("RZ", self.qubitsToList([self.__id, phi]))
        operator = super().RZ(self.__id, phi)
        self.setOperation(operator)

    # Two qubit gates
    # .......................................
    def CH(self, control):
        self.addOp("CH", self.qubitsToList([control, self.__id]))
        operator = super().CH(control, self.__id)
        self.setOperation(operator)

    def CX(self, control):
        self.addOp("CX", self.qubitsToList([control, self.__id]))
        operator = super().CX(control, self.__id)
        self.setOperation(operator)

    def CNOT(self, control):
        self.CX(control)

    def CY(self, control):
        self.addOp("CY", self.qubitsToList([control, self.__id]))
        operator = super().CY(control, self.__id)
        self.setOperation(operator)

    def CZ(self, control):
        self.addOp("CZ", self.qubitsToList([control, self.__id]))
        operator = super().CZ(control, self.__id)
        self.setOperation(operator)

    def CV(self, control, adjoint = False):
        self.addOp("CV", self.qubitsToList([control, self.__id]))
        operator = super().CV(control, self.__id, adjoint)
        self.setOperation(operator)

    def CS(self, control, adjoint = False):
        self.addOp("CS", self.qubitsToList([control, self.__id]))
        operator = super().CS(control, self.__id, adjoint)
        self.setOperation(operator)

    def CT(self, control, adjoint = False):
        self.addOp("CT", self.qubitsToList([control, self.__id]))
        operator = super().CT(control, self.__id, adjoint)
        self.setOperation(operator)

    def CRX(self, control, theta):
        self.addOp("CRX", self.qubitsToList([control, self.__id, theta]))
        operator = super().CRX(control, self.__id, theta)
        self.setOperation(operator)

    def CRY(self, control, theta):
        self.addOp("CRY", self.qubitsToList([control, self.__id, theta]))
        operator = super().CRY(control, self.__id, theta)
        self.setOperation(operator)

    def CRZ(self, control, phi):
        self.addOp("CRZ", self.qubitsToList([control, self.__id, phi]))
        operator = super().CRZ(control, self.__id, phi)
        self.setOperation(operator)

    def CU(self, control, *argv):
        self.addOp("CU", self.qubitsToList([control, self.__id]))
        operator = super().CU(control, self.__id, argv)
        self.setOperation(operator)

    def CU3(self, control, theta, phi, _lambda):
        self.addOp("CU3", self.qubitsToList([control, self.__id, theta, phi, _lambda]))
        operator = super().CU3(control, self.__id, theta, phi, _lambda)
        self.setOperation(operator)

    def CU2(self, control, phi, _lambda):
        self.addOp("CU2", self.qubitsToList([control, self.__id, phi, _lambda]))
        operator = super().CU2(control, self.__id, phi, _lambda)
        self.setOperation(operator)

    def CU1(self, control, _lambda):
        self.addOp("CU1", self.qubitsToList([control, self.__id, _lambda]))
        operator = super().CU1(control, self.__id, _lambda)
        self.setOperation(operator)

    def SWAP(self, target):
        self.addOp("SWAP", self.qubitsToList([self.__id, target]))
        operator = super().SWAP(self.__id, target)
        self.setOperation(operator)

    def Controlled(self, name, control, target, adjoint = False):
        self.addOp(name, self.qubitsToList([control, self.__id]))
        functions = {'X': super().CX, 'Y': super().CY, 'Z': super().CZ}
        operator = functions[name](control[0], control[1], self.__id)
        self.setOperation(operator)

    # Three qubit gates
    # .......................................
    def CCX(self, control1, control2):
        self.addOp("CCX", self.qubitsToList([control1, control2, self.__id]))
        operator = super().CCX(control1, control2, self.__id)
        self.setOperation(operator)

    def Toffoli(self, control1, control2):
        self.CCX(control1, control2)

    def Fredkin(self, target1, target2):
        self.addOp("Fredkin", self.qubitsToList([self.__id, target1, target2]))
        operator = super().Fredkin(self.__id, target1, target2)
        self.setOperation(operator)


"""
Registering qubits allows you to perform the same operation on several quibts at the same time.
In this class the quantum gates methods perform and record the type of operation.
"""
class QubitRegister(Qubit):
    def __init__(self, number = 3):
        self.__number = number
        if(self.getUsedQubits() < self.getQubitsNumber()):
            self.__reg = [Qubit() for i in range(1, number+1)]

    def __getitem__(self, key):
        return self.__reg[key-1]

    def getQubits(self):
        return self.__reg

    def X(self):
        for qubit in self.__reg:
            qubit.X()

    def Y(self):
        for qubit in self.__reg:
            qubit.Y()

    def Z(self):
        for qubit in self.__reg:
            qubit.Z()

    def V(self, target):
        for qubit in self.__reg:
            qubit.V()

    def S(self, target):
        for qubit in self.__reg:
            qubit.S()

    def T(self, target):
        for qubit in self.__reg:
            qubit.T()

    def H(self):
        for qubit in self.__reg:
            qubit.H()



#!/usr/bin/python
# -*- coding: UTF-8 -*-

# Author Cleoner S. Pietralonga
# e-mail: cleonerp@gmail.com
# Apache License

from cmath import *
from logicqubit.hilbert import *


"""
In this class, the numerical definition of operators is performed,
and the quantum gates methods performs the tensor product with the matrices in the correct order.
It is necessary to enter the qubit id as an input parameter.
"""
class Gates(Hilbert):

    def __init__(self, number_of_qubits=1):
        self.__number_of_qubits = number_of_qubits

    # basic matrices for the generation of operators
    # .......................................
    def ID(self):
        M = Matrix([[1, 0], [0, 1]], self.getCuda())
        return M

    def P0(self):
        M = Matrix([[1, 0], [0, 0]], self.getCuda())  # |0><0|
        return M

    def P1(self):
        M = Matrix([[0, 0], [0, 1]], self.getCuda())  # |1><1|
        return M

    def L0(self):
        M = Matrix([[0, 1], [0, 0]], self.getCuda())  # |0><1|
        return M

    def L1(self):
        M = Matrix([[0, 0], [1, 0]], self.getCuda())  # |1><0|
        return M

    # One qubit gates
    # input parameters: target
    # .......................................
    def X(self, target=1):
        M = Matrix([[0, 1], [1, 0]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def Y(self, target=1):
        M = Matrix([[0, -1j], [1j, 0]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def Z(self, target=1):
        M = Matrix([[1, 0], [0, -1]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def V(self, target=1, adjoint=False):
        M = Matrix([[1, -1j], [-1j, 1]], self.getCuda()) * ((1j + 1) / 2)  # sqrt(X) ou sqrt(NOT)
        if adjoint:
            M = M.adjoint()
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def S(self, target=1, adjoint=False):
        M = Matrix([[1, 0], [0, 1j]], self.getCuda())  # sqrt(Z)
        if adjoint:
            M = M.adjoint()
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def T(self, target=1, adjoint=False):
        M = Matrix([[1, 0], [0, (1 + 1j) / sqrt(2)]], self.getCuda())  # sqrt(S)
        if adjoint:
            M = M.adjoint()
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def H(self, target=1):
        M = Matrix([[1, 1], [1, -1]], self.getCuda()) * (1 / sqrt(2))
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def U(self, target, *argv):  # U or theta, phi and _lambda
        if len(argv) == 1:
            M = Matrix(argv[0][0], self.getCuda())
        else:
            theta = argv[0]
            phi = argv[1]
            _lambda = argv[2]
            M = Matrix(
                [[exp(-1j * (phi + _lambda) / 2) * cos(theta / 2), -exp(-1j * (phi - _lambda) / 2) * sin(theta / 2)],
                 [exp(-1j * (phi - _lambda) / 2) * sin(theta / 2), exp(1j * (phi + _lambda)) * cos(theta / 2)]],
                self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def U3(self, target, theta, phi, _lambda):
        M = Matrix([[cos(theta / 2), -exp(1j * _lambda) * sin(theta / 2)],
                    [exp(1j * phi) * sin(theta / 2), exp(1j * (phi + _lambda)) * cos(theta / 2)]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def U2(self, target, phi, _lambda):
        M = Matrix([[1, -exp(1j * _lambda)], [exp(1j * phi), exp(1j * (phi + _lambda))]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def U1(self, target, _lambda):
        M = Matrix([[1, 0], [0, exp(1j * _lambda)]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def RX(self, target, theta):
        M = Matrix([[cos(theta / 2), -1j * sin(theta / 2)],
                    [-1j * sin(theta / 2), cos(theta / 2)]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def RY(self, target, theta):
        M = Matrix([[cos(theta / 2), -sin(theta / 2)],
                    [sin(theta / 2), cos(theta / 2)]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    def RZ(self, target, phi):
        M = Matrix([[exp(-1j * phi / 2), 0], [0, exp(1j * phi / 2)]], self.getCuda())
        list = self.getOrdListSimpleGate(target, M)
        operator = self.kronProduct(list)
        return operator

    # Two qubit gates
    # input parameters: control and target
    # .......................................
    def CH(self, control, target):
        M = Matrix([[1, 1], [1, -1]], self.getCuda()) * (1 / sqrt(2)) # H
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator
    
    def CX(self, control, target):
        M = Matrix([[0, 1], [1, 0]], self.getCuda())  # X
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CNOT(self, control, target):
        return self.CX(control, target)

    def CY(self, control, target):
        M = Matrix([[0, -1j], [1j, 0]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CZ(self, control, target):
        M = Matrix([[1, 0], [0, -1]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CV(self, control, target, adjoint=False):
        M = Matrix([[1, -1j], [-1j, 1]], self.getCuda()) * ((1j + 1) / 2)  # sqrt(X) ou sqrt(NOT)
        if adjoint:
            M = M.adjoint()
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CS(self, control, target, adjoint=False):
        M = Matrix([[1, 0], [0, 1j]], self.getCuda())  # sqrt(Z)
        if adjoint:
            M = M.adjoint()
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CT(self, control, target, adjoint=False):
        M = Matrix([[1, 0], [0, (1 + 1j) / sqrt(2)]], self.getCuda())  # sqrt(S)
        if adjoint:
            M = M.adjoint()
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CRX(self, control, target, theta):
        M = Matrix([[cos(theta / 2), -1j * sin(theta / 2)],
                    [-1j * sin(theta / 2), cos(theta / 2)]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CRY(self, control, target, theta):
        M = Matrix([[cos(theta / 2), -sin(theta / 2)],
                    [sin(theta / 2), cos(theta / 2)]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CRZ(self, control, target, phi):
        M = Matrix([[exp(-1j * phi / 2), 0], [0, exp(1j * phi / 2)]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    # generic controlled gate
    def CU(self, control, target, *argv):  # U or theta, phi and _lambda
        if len(argv) == 1:
            M = Matrix(argv[0][0], self.getCuda())
        else:
            theta = argv[0]
            phi = argv[1]
            _lambda = argv[2]
            M = Matrix(
                [[exp(-1j * (phi + _lambda) / 2) * cos(theta / 2), -exp(-1j * (phi - _lambda) / 2) * sin(theta / 2)],
                 [exp(1j * (phi - _lambda) / 2) * sin(theta / 2), exp(1j * (phi + _lambda)) * cos(theta / 2)]],
                self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CU3(self, control, target, theta, phi, _lambda):
        M = Matrix([[cos(theta / 2), -exp(1j * _lambda) * sin(theta / 2)],
                    [exp(1j * phi) * sin(theta / 2), exp(1j * (phi + _lambda)) * cos(theta / 2)]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CU2(self, control, target, phi, _lambda):
        M = Matrix([[1, -exp(1j * _lambda)], [exp(1j * phi), exp(1j * (phi + _lambda))]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def CU1(self, control, target, _lambda):
        M = Matrix([[1, 0], [0, exp(1j * _lambda)]], self.getCuda())
        list1, list2 = self.getOrdListCtrlGate(control, target, M)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    # performs a state change of two qubits
    def SWAP(self, target1, target2):
        list1, list2, list3, list4 = self.getOrdListSWAP(target1, target2)
        operator = self.kronProduct(list1) + self.kronProduct(list2) + self.kronProduct(list3) + self.kronProduct(list4)
        return operator

    # Three qubit gates, which perform and record the operation
    # input parameters: control1, control2, and target
    # .......................................
    def CCX(self, control1, control2, target):
        Gate = Matrix([[0, 1], [1, 0]], self.getCuda()) - self.ID()
        list1, list2 = self.getOrdListCtrl2Gate(control1, control2, target, Gate)
        operator = self.kronProduct(list1) + self.kronProduct(list2)
        return operator

    def Toffoli(self, control1, control2, target):
        return self.CCX(control1, control2, target)

    # it's a controlled SWAP
    def Fredkin(self, control, target1, target2):
        list1, list2, list3, list4, list5, list6 = self.getOrdListFredkin(control, target1, target2)
        ID = self.kronProduct(list1)
        P1_SWAP = self.kronProduct(list2) + self.kronProduct(list3) + self.kronProduct(list4) + self.kronProduct(list5)
        P1_ID = self.kronProduct(list6)
        operator = ID + (P1_SWAP-P1_ID)
        return operator

    # orders the matrices for the tensor product of 1 qubit operations
    def getOrdListSimpleGate(self, target, Gate):
        list = []
        if self.isFirstLeft():
            plist = range(1, self.__number_of_qubits + 1)
        else:
            plist = reversed(range(1, self.__number_of_qubits + 1))
        for i in plist:
            if i == target:
                list.append(Gate)
            else:
                list.append(Matrix([[1, 0], [0, 1]], self.getCuda()))
        return list

    # orders the matrices for the tensor product of 2 qubits operations
    def getOrdListCtrlGate(self, control, target, Gate):
        list1 = []
        list2 = []
        if self.isFirstLeft():
            plist = range(1, self.__number_of_qubits + 1)
        else:
            plist = reversed(range(1, self.__number_of_qubits + 1))
        for i in plist:
            if i == control:
                list1.append(self.P0())  # |0><0|
                list2.append(self.P1())  # |1><1|
            elif i == target:
                list1.append(self.ID())
                list2.append(Gate)
            else:
                list1.append(self.ID())
                list2.append(self.ID())
        return list1, list2

    # orders the matrices for the tensor product of 3 qubits operations
    def getOrdListCtrl2Gate(self, control1, control2, target, Gate):
        list1 = []
        list2 = []
        if self.isFirstLeft():
            plist = range(1, self.__number_of_qubits + 1)
        else:
            plist = reversed(range(1, self.__number_of_qubits + 1))
        for i in plist:
            if i == control1 or i == control2:
                list1.append(self.ID())
                list2.append(self.P1())  # |1><1|
            elif i == target:
                list1.append(self.ID())
                list2.append(Gate)
            else:
                list1.append(self.ID())
                list2.append(self.ID())
        return list1, list2

    # orders the matrices for the tensor product of the SWAP gate operation
    def getOrdListSWAP(self, target1, target2):
        list1 = []
        list2 = []
        list3 = []
        list4 = []
        if self.isFirstLeft():
            plist = range(1, self.__number_of_qubits + 1)
        else:
            plist = reversed(range(1, self.__number_of_qubits + 1))
        for i in plist:
            if i == target1:
                list1.append(self.P0())  # |0><0|
                list2.append(self.L0())  # |0><1|
                list3.append(self.L1())  # |1><0|
                list4.append(self.P1())  # |1><1|
            elif i == target2:
                list1.append(self.P0())  # |0><0|
                list2.append(self.L1())  # |1><0|
                list3.append(self.L0())  # |0><1|
                list4.append(self.P1())  # |1><1|
            else:
                list1.append(self.ID())
                list2.append(self.ID())
                list3.append(self.ID())
                list4.append(self.ID())
        return list1, list2, list3, list4

    # orders the matrices for the tensor product of the Fredkin gate operation
    def getOrdListFredkin(self, control, target1, target2):
        list1 = []
        list2 = []
        list3 = []
        list4 = []
        list5 = []
        list6 = []
        if self.isFirstLeft():
            plist = range(1, self.__number_of_qubits + 1)
        else:
            plist = reversed(range(1, self.__number_of_qubits + 1))
        for i in plist:
            if i == control:
                list1.append(self.ID())  # ID
                list2.append(self.P1())  # SWAP P0xP0
                list3.append(self.P1())  # SWAP L0xL1
                list4.append(self.P1())  # SWAP L1xL0
                list5.append(self.P1())  # SWAP P1xP1
                list6.append(self.P1())  # -ID
            elif i == target1:
                list1.append(self.ID())
                list2.append(self.P0())  # |0><0|
                list3.append(self.L0())  # |0><1|
                list4.append(self.L1())  # |1><0|
                list5.append(self.P1())  # |1><1|
                list6.append(self.ID())
            elif i == target2:
                list1.append(self.ID())
                list2.append(self.P0())  # |0><0|
                list3.append(self.L1())  # |1><0|
                list4.append(self.L0())  # |0><1|
                list5.append(self.P1())  # |1><1|
                list6.append(self.ID())
            else:
                list1.append(self.ID())
                list2.append(self.ID())
                list3.append(self.ID())
                list4.append(self.ID())
                list5.append(self.ID())
                list6.append(self.ID())
        return list1, list2, list3, list4, list5, list6



#!/usr/bin/python
# -*- coding: UTF-8 -*-

# Author Cleoner S. Pietralonga
# e-mail: cleonerp@gmail.com
# Apache License

from logicqubit.utils import *
from logicqubit.hilbert import *

class PauliDecomposition:

    def __init__(self, H):
        self.H = Matrix(H)  # Hermitian matrix

    def sigma(self):
        ID = Matrix([[1, 0], [0, 1]])
        X = Matrix([[0, 1], [1, 0]])
        Y = Matrix([[0, -1j], [1j, 0]])
        Z = Matrix([[1, 0], [0, -1]])
        return [ID, X, Y, Z]

    def get_a(self):
        a = [[0]*4 for i in range(4)]
        for i, sigma_i in enumerate(self.sigma()):
            for j, sigma_j in enumerate(self.sigma()):
                a[i][j] = 1/4*(sigma_i.kron(sigma_j) * self.H).trace().get().item()
        return a